from pathlib import Path
from typing import Dict, List, Union

from synthesisai.data_types import OutOfFrameLandmarkStrategy
from synthesisai.item_loader import _ItemLoader
from synthesisai.item_loader_v1 import _ItemLoaderV1
from synthesisai.item_loader_v2 import _ItemLoaderV2
from synthesisai.modality import Modality


class _ItemLoaderFactory:
    @classmethod
    def _contains_v1_info(cls, root: Path) -> bool:
        for f in root.glob("*"):
            if str(f).endswith("info.json"):
                return True
        return False

    @classmethod
    def _contains_v2_info(cls, root: Path) -> bool:
        for f in root.glob("*"):
            if str(f).endswith("image_metadata.json"):
                return True
        return False

    @classmethod
    def get_item_loader(
        cls,
        root: Path,
        modalities: List[Modality],
        metadata_records: List[Dict[str, Union[int, str]]],
        out_of_frame_landmark_strategy: OutOfFrameLandmarkStrategy,
        body_segmentation_mapping: Dict[str, int],
        clothing_segmentation_mapping: Dict[str, int],
        face_segmentation_classes: List[str],
        face_bbox_pad: int,
    ) -> _ItemLoader:
        contains_v1_info = cls._contains_v1_info(root)
        contains_v2_info = cls._contains_v2_info(root)

        if not contains_v1_info and not contains_v2_info:
            raise Exception(
                f"Neither *.info.json nor *.image_metadata.json files were found in {root}"
            )

        if contains_v1_info and contains_v2_info:
            raise Exception(
                f"Both *.info.json and *.image_metadata.json files were found in {root}"
            )

        if contains_v1_info:
            item_loader_cls = _ItemLoaderV1
        elif contains_v2_info:
            item_loader_cls = _ItemLoaderV2
    
        return item_loader_cls(
            root,
            modalities,
            metadata_records,
            out_of_frame_landmark_strategy,
            body_segmentation_mapping,
            clothing_segmentation_mapping,
            face_segmentation_classes,
            face_bbox_pad,
        )



import enum
import json
import logging
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd
import tiffile

from synthesisai.data_types import (
    InstanceId,
    Item,
    Landmark_2D,
    Landmark_3D,
    LandmarkId,
    OutOfFrameLandmarkStrategy,
)
from synthesisai.item_loader import N_IBUG68_LANDMARKS, _ItemLoader
from synthesisai.modality import Modality

logger = logging.getLogger(__name__)


class _Extension(str, Enum):
    INFO = "info.json"
    RGB = "rgb.png"
    NORMALS = "normals.tif"
    DEPTH = "depth.tif"
    SEGMENTS = "segments.png"
    MEDIAPIPE_DENSE_OBJ = "mediapipe_dense.obj"
    SAI_DENSE_OBJ = "sai_dense.obj"


def _modality_files(modality: Modality) -> List[_Extension]:
    return {
        Modality.SCENE_ID: [_Extension.INFO],
        Modality.RGB: [_Extension.RGB],
        Modality.NORMALS: [_Extension.NORMALS],
        Modality.DEPTH: [_Extension.DEPTH],
        Modality.BODY_SEGMENTATION: [_Extension.INFO, _Extension.SEGMENTS],
        Modality.CLOTHING_SEGMENTATION: None,
        Modality.INSTANCE_SEGMENTATION: [_Extension.INFO, _Extension.SEGMENTS],
        Modality.LANDMARKS_IBUG68: [_Extension.INFO],
        Modality.LANDMARKS_CONTOUR_IBUG68: [_Extension.INFO],
        Modality.LANDMARKS_KINECT_V2: [_Extension.INFO],
        Modality.LANDMARKS_COCO: [_Extension.INFO],
        Modality.LANDMARKS_MEDIAPIPE: [_Extension.INFO],
        Modality.LANDMARKS_MPEG4: [_Extension.INFO],
        Modality.LANDMARKS_3D_IBUG68: [_Extension.INFO],
        Modality.LANDMARKS_3D_KINECT_V2: [_Extension.INFO],
        Modality.LANDMARKS_3D_COCO: [_Extension.INFO],
        Modality.LANDMARKS_3D_MEDIAPIPE: [_Extension.INFO],
        Modality.LANDMARKS_3D_MPEG4: [_Extension.INFO],
        Modality.PUPILS: [_Extension.INFO],
        Modality.PUPILS_3D: [_Extension.INFO],
        Modality.IDENTITY: [_Extension.INFO],
        Modality.IDENTITY_METADATA: [_Extension.INFO],
        Modality.HAIR: [_Extension.INFO],
        Modality.FACIAL_HAIR: [_Extension.INFO],
        Modality.EXPRESSION: [_Extension.INFO],
        Modality.GAZE: [_Extension.INFO],
        Modality.FACE_BBOX: [_Extension.RGB, _Extension.INFO, _Extension.SEGMENTS],
        Modality.CAM_TO_HEAD: [_Extension.INFO],
        Modality.HEAD_TO_CAM: [_Extension.INFO],
        Modality.WORLD_TO_HEAD: [_Extension.INFO],
        Modality.HEAD_TO_WORLD: [_Extension.INFO],
        Modality.CAM_TO_WORLD: [_Extension.INFO],
        Modality.WORLD_TO_CAM: [_Extension.INFO],
        Modality.CAM_INTRINSICS: [_Extension.INFO],
        Modality.CAMERA_NAME: [_Extension.INFO],
        Modality.FRAME_NUM: [_Extension.INFO],
        Modality.LANDMARKS_3D_MEDIAPIPE_FACE: [
            _Extension.MEDIAPIPE_DENSE_OBJ,
            _Extension.INFO,
        ],
        Modality.LANDMARKS_3D_SAI: [_Extension.SAI_DENSE_OBJ, _Extension.INFO],
        Modality.LANDMARKS_MEDIAPIPE_FACE: [
            _Extension.MEDIAPIPE_DENSE_OBJ,
            _Extension.INFO,
        ],
        Modality.LANDMARKS_SAI: [_Extension.SAI_DENSE_OBJ, _Extension.INFO],
    }[modality]


class _ItemLoaderV1(_ItemLoader):
    def __init__(
        self,
        root: Path,
        modalities: List[Modality],
        metadata_records: List[Dict[str, Union[int, str]]],
        out_of_frame_landmark_strategy: OutOfFrameLandmarkStrategy,
        body_segmentation_mapping: Dict[str, int],
        clothing_segmentation_mapping: Dict[str, int],
        face_segmentation_classes: List[str],
        face_bbox_pad: int,
    ):
        super().__init__(
            root,
            modalities,
            metadata_records,
            out_of_frame_landmark_strategy,
            body_segmentation_mapping,
            clothing_segmentation_mapping,
            face_segmentation_classes,
            face_bbox_pad,
        )
        filtered_modalities = []
        for m in self._modalities:
            if _modality_files(m) is None:
                logger.warn(f"Ignoring unsupported modality {m}")
            else:
                filtered_modalities.append(m)
        self._modalities = filtered_modalities
        self._needs_info = False
        for modality in self._modalities:
            if _Extension.INFO in _modality_files(modality):
                self._needs_info = True

        for record in metadata_records:
            scene_id = record["SCENE_ID"]
            camera = record["CAMERA"]
            frame = record["FRAME"]

            for modality in self._modalities:
                for extension in _modality_files(modality):
                    expected_file = (
                        root / f"{scene_id}.{camera}.{frame}.{extension.value}"
                    )
                    if not expected_file.exists():
                        raise ValueError(
                            f"Can't find file '{expected_file}' required for {modality.name} modality."
                        )

    def get_item(
        self, element_idx: Tuple[int, str, int], item_meta: pd.DataFrame
    ) -> Item:
        info = None
        if self._needs_info:
            info_file = Path(
                item_meta[item_meta.EXTENSION == _Extension.INFO].file_path.iloc[0]
            )
            with info_file.open("r") as f:
                info = json.load(f)

        ret: Item = {}
        for modality in self._modalities:
            ret[modality] = self._open_modality(modality, item_meta, element_idx, info)
        return ret

    def _open_modality(
        self,
        modality: Modality,
        item_meta: pd.DataFrame,
        element_idx: Tuple[int, str, int],  # (scene id, camera name, frame number)
        info: Optional[dict],
    ) -> Any:
        scene_id = item_meta.SCENE_ID.iloc[0]
        if modality == Modality.SCENE_ID:
            return scene_id

        if modality == Modality.RGB:
            file_path = item_meta[item_meta.EXTENSION == _Extension.RGB].file_path.iloc[
                0
            ]
            return self._read_rgb(file_path, element_idx)

        if modality == Modality.BODY_SEGMENTATION:
            file_path = item_meta[
                item_meta.EXTENSION == _Extension.SEGMENTS
            ].file_path.iloc[0]
            segment_img, _ = self._read_body_segmentation(file_path, element_idx, info)
            return segment_img

        if modality == Modality.INSTANCE_SEGMENTATION:
            file_path = item_meta[
                item_meta.EXTENSION == _Extension.SEGMENTS
            ].file_path.iloc[0]
            segment_img = self._read_instance_segmentation(file_path, element_idx, info)
            return segment_img

        if modality == Modality.NORMALS:
            normals_file = item_meta[
                item_meta.EXTENSION == _Extension.NORMALS
            ].file_path.iloc[0]
            img = tiffile.imread(str(normals_file))
            if img is None:
                raise ValueError(f"Error reading {normals_file}")
            if element_idx in self._image_sizes:
                if self._image_sizes[element_idx] != img.shape[1::-1]:
                    raise ValueError(
                        f"Dimensions of different image modalities do not match for SCENE_ID={scene_id}"
                    )
            else:
                self._image_sizes[element_idx] = img.shape[1::-1]
            return img

        if modality == Modality.DEPTH:
            depth_file = item_meta[
                item_meta.EXTENSION == _Extension.DEPTH
            ].file_path.iloc[0]
            img = tiffile.imread(str(depth_file))
            if element_idx in self._image_sizes:
                if self._image_sizes[element_idx] != img.shape[::-1]:
                    raise ValueError(
                        f"Dimensions of different image modalities do not match for SCENE_ID={scene_id}"
                    )
            else:
                self._image_sizes[element_idx] = img.shape[::-1]
            if img is None:
                raise ValueError(f"Error reading {depth_file}")
            return img

        if modality in (Modality.LANDMARKS_IBUG68, Modality.LANDMARKS_CONTOUR_IBUG68):
            return self._read_face_landmarks_2d(info, modality, element_idx)

        if modality in (
            Modality.LANDMARKS_KINECT_V2,
            Modality.LANDMARKS_COCO,
            Modality.LANDMARKS_MEDIAPIPE,
            Modality.LANDMARKS_MPEG4,
        ):
            landmark_meta = self._read_body_landmarks(
                info, modality, self._image_sizes[element_idx]
            )
            return landmark_meta

        if modality in (
            Modality.LANDMARKS_3D_IBUG68,
            Modality.LANDMARKS_3D_KINECT_V2,
            Modality.LANDMARKS_3D_COCO,
            Modality.LANDMARKS_3D_MEDIAPIPE,
            Modality.LANDMARKS_3D_MPEG4,
        ):
            landmark_meta = self._read_landmarks_3d(info, modality)
            return landmark_meta

        if modality == Modality.PUPILS:
            pupils = [
                info["pupil_coordinates"]["pupil_left"]["screen_space_pos"],
                info["pupil_coordinates"]["pupil_right"]["screen_space_pos"],
            ]
            if element_idx not in self._image_sizes:
                raise ValueError(
                    "Pupils can only be loaded with at least one image modality"
                )
            scale = self._image_sizes[element_idx]
            pupils_np = np.array(pupils, dtype=np.float64) * scale
            pupils = {
                "pupil_left": tuple(pupils_np[0]),
                "pupil_right": tuple(pupils_np[1]),
            }

            # v1 only has one character with ID 1
            pupils = {1: pupils}
            return pupils

        if modality == Modality.PUPILS_3D:
            pupils = {
                "pupil_left": tuple(
                    info["pupil_coordinates"]["pupil_left"]["camera_space_pos"]
                ),
                "pupil_right": tuple(
                    info["pupil_coordinates"]["pupil_right"]["camera_space_pos"]
                ),
            }

            # v1 only has one character with ID 1
            pupils = {1: pupils}
            return pupils

        if modality == Modality.IDENTITY:
            identity = info["identity_metadata"]["id"]
            # v1 only has one character with ID 1
            identity = {1: identity}
            return identity

        if modality == Modality.IDENTITY_METADATA:
            metadata = info["identity_metadata"]
            metadata = {1: metadata}
            return metadata

        if modality == Modality.HAIR:
            if "hair" in info["facial_attributes"]:
                hair = info["facial_attributes"]["hair"]
            else:
                hair = None

            # v1 only has one character with ID 1
            hair = {1: hair}
            return hair

        if modality == Modality.FACIAL_HAIR:
            if "facial_hair" in info["facial_attributes"]:
                facial_hair = info["facial_attributes"]["facial_hair"]
            else:
                facial_hair = None

            # v1 only has one character with ID 1
            facial_hair = {1: facial_hair}
            return facial_hair

        if modality == Modality.EXPRESSION:
            expression = info["facial_attributes"]["expression"]

            # v1 only has one character with ID 1
            expression = {1: expression}
            return expression

        if modality == Modality.GAZE:
            gaze = {
                "right": info["gaze_values"]["eye_right"]["gaze_vector"],
                "left": info["gaze_values"]["eye_left"]["gaze_vector"],
            }

            # v1 only has one character with ID 1
            gaze = {1: gaze}
            return gaze

        if modality == Modality.FACE_BBOX:
            file_path = item_meta[
                item_meta.EXTENSION == _Extension.SEGMENTS
            ].file_path.iloc[0]
            segment_img, segment_mapping_int = self._read_body_segmentation(
                file_path, element_idx, info
            )

            segment_mapping = info["segments_mapping"]
            face_segments = set(self._face_segmentation_classes) & set(
                segment_mapping.keys()
            )
            face_seg_idxs = [
                segment_mapping_int[segment_mapping[s]] for s in face_segments
            ]
            face_mask = np.isin(segment_img, face_seg_idxs).astype(np.uint16)

            def get_bbox(img: np.ndarray):
                yxs = np.where(img != 0)
                if len(yxs[0]) == 0:
                    return (-1, -1, -1, -1)
                bbox = np.min(yxs[1]), np.min(yxs[0]), np.max(yxs[1]), np.max(yxs[0])
                return bbox

            height, width = self._image_sizes[element_idx]

            def expand_bbox(bbox: Tuple[int, int, int, int], padding: int):
                x0, y0, x1, y1 = bbox
                x0 = max(0, x0 - padding)
                y0 = max(0, y0 - padding)
                x1 = min(x1 + padding, width)
                y1 = min(y1 + padding, height)

                return (x0, y0, x1, y1)

            face_bbox = get_bbox(face_mask)
            expanded_face_bbox = expand_bbox(face_bbox, self._face_bbox_pad)

            # v1 only has one character with ID 1
            expanded_face_bbox = {1: expanded_face_bbox}

            return expanded_face_bbox

        if modality == Modality.CAM_TO_WORLD:
            return np.array(
                info["camera"]["transform_cam2world"]["mat_4x4"], dtype=np.float64
            )

        if modality == Modality.WORLD_TO_CAM:
            return np.array(
                info["camera"]["transform_world2cam"]["mat_4x4"], dtype=np.float64
            )

        if modality == Modality.HEAD_TO_CAM:
            head2cam = np.array(
                info["head_transform"]["transform_head2cam"]["mat_4x4"],
                dtype=np.float64,
            )

            # v1 only has one character with ID 1
            head2cam = {1: head2cam}
            return head2cam

        if modality == Modality.CAM_TO_HEAD:
            head2cam = self._open_modality(
                Modality.HEAD_TO_CAM, item_meta, element_idx, info
            )
            cam2head = {}
            for instance_id, h2c in head2cam.items():
                cam2head[instance_id] = np.linalg.inv(h2c)

            return cam2head

        if modality == Modality.HEAD_TO_WORLD:
            head2world = np.array(
                info["head_transform"]["transform_head2world"]["mat_4x4"],
                dtype=np.float64,
            )

            # v1 only has one character with ID 1
            head2world = {1: head2world}
            return head2world

        if modality == Modality.WORLD_TO_HEAD:
            head2world = self._open_modality(
                Modality.HEAD_TO_WORLD, item_meta, element_idx, info
            )
            world2head = {}
            for instance_id, h2w in head2world.items():
                world2head[instance_id] = np.linalg.inv(h2w)

            return world2head

        if modality == Modality.CAM_INTRINSICS:
            return np.array(info["camera"]["intrinsics"], dtype=np.float64)

        if modality == Modality.CAMERA_NAME:
            return item_meta.CAMERA_NAME.iloc[0]

        if modality == Modality.FRAME_NUM:
            return item_meta.FRAME_NUM.iloc[0]

        if modality == Modality.LANDMARKS_3D_MEDIAPIPE_FACE:
            m = np.array(
                info["camera"]["transform_world2cam"]["mat_4x4"], dtype=np.float64
            )
            file_path = item_meta[
                item_meta.EXTENSION == _Extension.MEDIAPIPE_DENSE_OBJ
            ].file_path.iloc[0]
            return self._read_obj(file_path, m)

        if modality == Modality.LANDMARKS_3D_SAI:
            m = np.array(
                info["camera"]["transform_world2cam"]["mat_4x4"], dtype=np.float64
            )
            file_path = item_meta[
                item_meta.EXTENSION == _Extension.SAI_DENSE_OBJ
            ].file_path.iloc[0]
            return self._read_obj(file_path, m)

        if modality == Modality.LANDMARKS_SAI:
            sai_3d = self._open_modality(
                Modality.LANDMARKS_3D_SAI, item_meta, element_idx, info
            )
            intrinsics = self._open_modality(
                Modality.CAM_INTRINSICS, item_meta, element_idx, info
            )

            landmarks_sai = {}
            for instance_id, landmarks in sai_3d.items():
                landmarks_sai[instance_id] = dict(
                    enumerate(
                        self._hom_to_euclidian(
                            np.tensordot(
                                landmarks * [1, -1, -1], intrinsics, axes=(-1, 1)
                            )
                        )
                    )
                )
            return landmarks_sai

        if modality == Modality.LANDMARKS_MEDIAPIPE_FACE:
            mediapipe_face_3d = self._open_modality(
                Modality.LANDMARKS_3D_MEDIAPIPE_FACE, item_meta, element_idx, info
            )
            intrinsics = self._open_modality(
                Modality.CAM_INTRINSICS, item_meta, element_idx, info
            )

            face_landmarks = {}
            for instance_id, landmarks in mediapipe_face_3d.items():
                face_landmarks[instance_id] = dict(
                    enumerate(
                        self._hom_to_euclidian(
                            np.tensordot(
                                landmarks * [1, -1, -1], intrinsics, axes=(-1, 1)
                            )
                        )
                    )
                )
            return face_landmarks

        raise ValueError("Unknown modality")

    def _read_body_segmentation(
        self, body_segmentation_file: str, element_idx: tuple, info: Dict[str, Any]
    ) -> Tuple[np.ndarray, np.ndarray]:
        img = cv2.imread(str(body_segmentation_file), cv2.IMREAD_UNCHANGED)
        if img is None:
            raise ValueError(f"Error reading {body_segmentation_file}")
        if element_idx in self._image_sizes:
            if self._image_sizes[element_idx] != img.shape[::-1]:
                raise ValueError(
                    f"Dimensions of different image modalities do not match."
                )
        else:
            self._image_sizes[element_idx] = img.shape[::-1]

        body_segmentation_mapping = info["segments_mapping"]
        body_segmentation_mapping_int = np.full(
            np.max(list(body_segmentation_mapping.values())) + 1,
            self._body_segmentation_mapping["default"],
            dtype=np.uint8,
        )
        for key, value in body_segmentation_mapping.items():
            if key in self._body_segmentation_mapping:
                body_segmentation_mapping_int[value] = self._body_segmentation_mapping[
                    key
                ]

        segment_img = body_segmentation_mapping_int[img]
        return segment_img, body_segmentation_mapping_int

    def _read_instance_segmentation(
        self, body_segmentation_file: str, element_idx: tuple, info: Dict[str, Any]
    ) -> np.ndarray:
        img = cv2.imread(str(body_segmentation_file), cv2.IMREAD_UNCHANGED)
        if img is None:
            raise ValueError(f"Error reading {body_segmentation_file}")
        if element_idx in self._image_sizes:
            if self._image_sizes[element_idx] != img.shape[::-1]:
                raise ValueError(
                    f"Dimensions of different image modalities do not match."
                )
        else:
            self._image_sizes[element_idx] = img.shape[::-1]

        segment_mapping = info["segments_mapping"]
        assert "background" in segment_mapping

        background_index = segment_mapping["background"]
        instance_img = img.copy()
        # there is only one character in v1, which has index 0
        instance_img[img == background_index] = 0
        instance_img[img != background_index] = 1

        return instance_img

    def _read_rgb(self, rgb_file: str, element_idx: tuple) -> np.ndarray:
        img = cv2.imread(str(rgb_file), cv2.IMREAD_COLOR)
        if img is None:
            raise ValueError(f"Error reading {rgb_file}")
        if element_idx in self._image_sizes:
            if self._image_sizes[element_idx] != img.shape[1::-1]:
                raise ValueError("Dimensions of different modalities do not match")
        else:
            self._image_sizes[element_idx] = img.shape[1::-1]
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        return img

    @staticmethod
    def _read_modality_meta(info: dict, mdt: Modality) -> dict:
        if mdt in (Modality.LANDMARKS_IBUG68, Modality.LANDMARKS_3D_IBUG68):
            meta_dict = info["landmarks"]["ibug68"]
        elif mdt is Modality.LANDMARKS_CONTOUR_IBUG68:
            meta_dict = info["landmarks"]["ibug68_contour"]
        elif mdt in (Modality.LANDMARKS_COCO, Modality.LANDMARKS_3D_COCO):
            meta_dict = info["landmarks"]["coco"]["whole_body"]
        elif mdt in (Modality.LANDMARKS_KINECT_V2, Modality.LANDMARKS_3D_KINECT_V2):
            meta_dict = info["landmarks"]["kinect_v2"]
        elif mdt in (Modality.LANDMARKS_MEDIAPIPE, Modality.LANDMARKS_3D_MEDIAPIPE):
            meta_dict = info["landmarks"]["mediapipe"]["body"]
        elif mdt in (Modality.LANDMARKS_MPEG4, Modality.LANDMARKS_3D_MPEG4):
            meta_dict = info["landmarks"]["mpeg4"]
        else:
            raise ValueError(f"Unrecognized modality for 3D landmarks {mdt}.")
        return meta_dict

    @classmethod
    def _read_body_landmarks(
        cls, info: dict, mdt: Modality, image_size: Tuple[int, int]
    ) -> Dict[InstanceId, Dict[LandmarkId, Landmark_2D]]:
        meta_dict = cls._read_modality_meta(info, mdt)
        result = {}

        for item in meta_dict:
            idd = item["id"]
            x, y = item["screen_space_pos"]
            w, h = image_size
            result[int(idd)] = (x * w, y * h)

        # v1 only has one character with ID 1
        result = {1: result}

        return result

    def _read_face_landmarks_2d(
        self, info: dict, mdt: Modality, element_idx: tuple
    ) -> Dict[InstanceId, Dict[LandmarkId, Landmark_2D]]:
        if element_idx not in self._image_sizes:
            raise ValueError(
                "Landmarks can only be loaded with at least one image modality"
            )

        meta_dict = self._read_modality_meta(info, mdt)

        if mdt is Modality.LANDMARKS_IBUG68:
            n_landmarks = len(meta_dict)
            if n_landmarks != N_IBUG68_LANDMARKS:
                msg = f"Error reading landmarks for item. Got {n_landmarks} landmarks instead of {N_IBUG68_LANDMARKS}"
                raise ValueError(msg)

        w, h = self._image_sizes[element_idx]
        landmarks: Dict[LandmarkId, Landmark_2D] = {}

        for landmark in meta_dict:
            x, y = landmark["screen_space_pos"]
            x, y = x * w, y * h
            landmarks[int(landmark["id"])] = (x, y)

        if self._out_of_frame_landmark_strategy is OutOfFrameLandmarkStrategy.CLIP:
            landmarks = OutOfFrameLandmarkStrategy.clip_landmarks_(landmarks, h, w)

        # v1 only has one character with ID 1
        landmarks = {1: landmarks}
        return landmarks

    @classmethod
    def _read_landmarks_3d(
        cls, info: dict, mdt: Modality
    ) -> Dict[LandmarkId, Landmark_3D]:
        meta_dict = cls._read_modality_meta(info, mdt)
        landmarks: Dict[LandmarkId, Landmark_3D] = {}

        for landmark in meta_dict:
            lmk_name = landmark["id"]
            landmarks[int(lmk_name)] = tuple(landmark["camera_space_pos"])

        # v1 only has one character with ID 1
        landmarks = {1: landmarks}
        return landmarks

    def _read_obj(self, obj_file: str, world_to_cam: np.ndarray):
        with open(obj_file, "r") as f:
            lines = f.readlines()
            result = np.empty((0, 3))
            for line in lines:
                if line.startswith("v "):
                    _, x, y, z = line.split()
                    x, y, z = float(x), float(y), float(z)
                    result = np.vstack((result, [x, y, z]))

        result = (
            np.tensordot(result, world_to_cam[:3, :3], axes=(-1, 1))
            + world_to_cam[:3, 3]
        )

        # v1 only has one character with ID 1
        result = {1: result}
        return result



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np

from synthesisai.data_types import OutOfFrameLandmarkStrategy
from synthesisai.modality import Modality

N_IBUG68_LANDMARKS = 68


class _ItemLoader(ABC):
    def __init__(
        self,
        root: Path,
        modalities: List[Modality],
        metadata_records: List[Dict[str, Union[int, str]]],
        out_of_frame_landmark_strategy: OutOfFrameLandmarkStrategy,
        body_segmentation_mapping: Dict[str, int],
        clothing_segmentation_mapping: Dict[str, int],
        face_segmentation_classes: List[str],
        face_bbox_pad: int,
    ):
        self._modalities = modalities
        self._out_of_frame_landmark_strategy = out_of_frame_landmark_strategy
        self._body_segmentation_mapping = body_segmentation_mapping
        self._clothing_segmentation_mapping = clothing_segmentation_mapping
        self._face_segmentation_classes = face_segmentation_classes
        self._face_bbox_pad = face_bbox_pad
        self._image_sizes: Dict[tuple, Tuple[int, int]] = {}

    @abstractmethod
    def get_item(self):
        return NotImplemented

    @staticmethod
    def _hom_to_euclidian(x):
        return x[:, :2] / np.expand_dims(x[:, 2], -1)

    @classmethod
    def get_euler_angles(
        cls, matrix: np.ndarray, order: str, degrees: bool
    ) -> np.ndarray:
        """
        Euler angles by matrix 4x4.

        :param np.ndarray matrix:  matrix 4x4.
        :param str order:  axes order ('xyz','yzx','zxy').
        :param bool degrees:  whether use degrees or radians.

        """
        from scipy.spatial.transform.rotation import Rotation

        return Rotation.from_matrix(matrix[:3, :3]).as_euler(order, degrees=degrees)

    @classmethod
    def get_quaternion(cls, matrix: np.ndarray) -> np.ndarray:
        """
        Quaternion by matrix 4x4.

        :param np.ndarray matrix:  matrix 4x4

        """
        from scipy.spatial.transform.rotation import Rotation

        return Rotation.from_matrix(matrix[:3, :3]).as_quat()

    @classmethod
    def get_shift(cls, matrix: np.ndarray) -> np.ndarray:
        """
        Return shift vector by matrix 4x4 which contains both rotation and shift.

        :param np.ndarray matrix:  matrix 4x4

        """
        return matrix[:3, 3]



from enum import Enum, auto


class Modality(Enum):
    """
    Different modalities of Synthesis AI dataset.
    All image modalities are in `[y][x][channel]` format, with axis going as follows::

        ┌-----> x
        |
        |
        v
        y
    """

    SCENE_ID = auto()
    """
    Scene ID (rendered scene number). 
    
    **Type**: `int`.
    """
    RGB = auto()
    """
    RGB image modality. 
    
    **Type**: `ndarray[uint8]`. **Channels**: `3`.
    """
    NORMALS = auto()
    """
    Normals image. All values are in [-1,1] range.
    
    **Type**: `ndarray[float16]`. **Channels**: 3.
    """
    DEPTH = auto()
    """
    Depth Image. All values are positive floats. Background has depth=0.
    
    **Type**: `ndarray[float16]`. **Channels**: 1.
    """
    BODY_SEGMENTATION = auto()
    """
    Semantic segmentation map of various body parts.
    
    **Type**: `ndarray[uint16]`. **Channels**: 1.
    """
    CLOTHING_SEGMENTATION = auto()
    """
    Semantic segmentation map of different types of clothing.
    
    **Type**: `ndarray[uint16]`. **Channels**: 1.
    """
    INSTANCE_SEGMENTATION = auto()
    """
    Semantic segmentation map defining the various characters in the image.
    
    **Type**: `ndarray[uint16]`. **Channels**: 1.
    """
    LANDMARKS_IBUG68 = auto()
    """
    iBUG-68 landmarks. Each landmark is given by name and two coordinates (x,y) in pixels.
    Each keypoint is a 2D projection of a 3D landmark.
    
    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark2d]`. Should have no more than 68 points.
    """
    LANDMARKS_CONTOUR_IBUG68 = auto()
    """
    iBUG-68 contour landmarks. Each landmark is given by two coordinates (name, x,y) in pixels.
    Each keypoint is defined in a similar manner to human labelers marking 2D face kepoints.
    
    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark2d]]`. Should have no more than 68 points.
    """
    LANDMARKS_KINECT_V2 = auto()
    """
    Kinect v2 landmarks. Each landmark by name and two coordinates (x,y) in pixels.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark2d]]`. Should have no more than 32 points.
    """
    LANDMARKS_MEDIAPIPE = auto()
    """
    MediaPipe pose landmarks. Each landmark is given by name and two coordinates (x,y) in pixels.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark2d]]`. Should have no more than 33 points.
    """
    LANDMARKS_COCO = auto()
    """
    COCO whole body landmarks. Each landmark is given by name and two coordinates (x,y) in pixels.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark2d]]`. Should have no more than 133 points.
    """
    LANDMARKS_MPEG4 = auto()
    """
    MPEG4 landmarks. Each landmark is given by name and two coordinates (x,y) in pixels.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark2d]]`.
    """
    LANDMARKS_3D_IBUG68 = auto()
    """
    iBUG-68 landmarks in 3D. Each landmark is given by name and three coordinates (x,y,z) in camera space.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark3d]]`. Should have no more than 68 points.
    """
    LANDMARKS_3D_KINECT_V2 = auto()
    """
    Kinect v2 landmarks in 3D. Each landmark is given by name and three coordinates (x,y,z) in camera space.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark3d]]`. Should have no more than 32 points.
    """
    LANDMARKS_3D_MEDIAPIPE = auto()
    """
    MediaPipe pose landmarks in 3D. Each landmark is given by name and three coordinates (x,y,z) in camera space.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark3d]]`. hould have no more than 33 points.
    """
    LANDMARKS_3D_COCO = auto()
    """
    COCO whole body landmarks in 3D. Each landmark is given by name and three coordinates (x,y,z) in camera space.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark3d]]`. Should have no more than 133 points.
    """
    LANDMARKS_3D_MPEG4 = auto()
    """
    MPEG4 landmarks in 3D. Each landmark is given by name and three coordinates (x,y,z) in camera space.

    **Type**: `Dict[InstanceId, Dict[LandmarkId, Landmark3d]]`.
    """
    PUPILS = auto()
    """
    Coordinates of pupils. Each pupil is given by name and two coordinates (x,y) in pixels.
    
    **Type**: `Dict[InstanceId, Dict[str, Landmark2d]]`.
    """
    PUPILS_3D = auto()
    """
    Coordinates of pupils in 3D. Each pupil is given by name and three coordinates (x,y,z) in camera space.
 
    **Type**: `Dict[InstanceId, Dict[str, Landmark3d]]`.
    """
    IDENTITY = auto()
    """
    Mapping from instance ID (ranging from 0 to the number of humans in the image) to human ID (corresponding to the ID provided in the job config)
    
    **Type**: `Dict[InstanceId, int]`.
    """
    IDENTITY_METADATA = auto()
    """
    Additional metadata about the people in the image.
    
    **Format**::
    
        {
            instance_id: {
                'gender': 'female'|'male',
                'age': int,
                'weight_kg': int,
                'height_cm': int,
                'id': int,
                'ethnicity': 'arab'|'asian'|'black'|'hisp'|'white'
            },
            ...
        }
    """
    HAIR = auto()
    """
    Hair metadata. If no hair are present `None` is returned.
    
    **Format**::
    
        {
            instance_id: {
                'relative_length': float64,
                'relative_density': float64,
                'style': str,
                'color_seed': float64,
                'color': str
            },
            ...
        }
    """
    FACIAL_HAIR = auto()
    """
    Facial hair metadata. If no facial hair is present for a human, `None` is provided.

    **Format**::
    
        {
            instance_id: {
                'relative_length': float64,
                'relative_density': float64,
                'style': str,
                'color_seed': float64,
                'color': str
            },
            ...
        }
    """
    EXPRESSION = auto()
    """
    Expression and its intensity.
    
    **Format**::
    
        {
            instance_id: {
                'intensity': float64, 
                'name': str
            },
            ...
        }
    """
    GAZE = auto()
    """
    Gaze direction in camera space.
    
    **Format**::
        {
            instance_id: {
                'horizontal_angle': ndarray[float64] **Shape**: `(3,)`.
                'vertical_angle': ndarray[float64] **Shape**: `(3,)`.
            },
            ...
        }
    """
    FACE_BBOX = auto()
    """
    Face bounding box in the format (left, top, right, bottom) in pixels.
    
    **Type**: `Dict[InstanceId, Tuple[int, int, int, int]]`.
    """
    HEAD_TO_CAM = auto()
    """
    Transformation matrix from the head to the camera coordinate system.
    
    **Type**: `Dict[InstanceId, ndarray[float32]]`. **Shape**: `(4, 4)`.
    """
    CAM_TO_HEAD = auto()
    """  
    Transformation matrix from the camera to the head coordinate system.
    
    **Type**: `Dict[InstanceId, ndarray[float32]]`. **Shape**: `(4, 4)`.
    """
    HEAD_TO_WORLD = auto()
    """
    Transformation matrix from the head to the world coordinate system.
    
    **Type**: `Dict[InstanceId, ndarray[float32]]`. **Shape**: `(4, 4)`.
    """
    WORLD_TO_HEAD = auto()
    """
    Transformation matrix from the world to the head coordinate system.
    
    **Type**: `Dict[InstanceId, ndarray[float32]]`. **Shape**: `(4, 4)`.
    """
    CAM_TO_WORLD = auto()
    """
    Transformation matrix from the camera to the world coordinate system.
    
    **Type**: `ndarray[float32]`. **Shape**: `(4, 4)`.
    """
    WORLD_TO_CAM = auto()
    """
    Transformation matrix from the world to the camera coordinate system.
    
    **Type**: `ndarray[float32]`. **Shape**: `(4, 4)`.
    """
    CAM_INTRINSICS = auto()
    """
    Camera intrinsics matrix in OpenCV format: https://docs.opencv.org/3.4.15/dc/dbb/tutorial_py_calibration.html.

    **Type**: `ndarray[float32]`. **Shape**: `(4, 4)`.
    """
    CAMERA_NAME = auto()
    """
    Camera name consisting of lowercase alphanumeric characters. Usually used when more than one are defined in a scene.
    Default is "cam_default".
    
    **Type**: `str`. 
    """
    FRAME_NUM = auto()
    """
    Frame number used for consecutive animation frames. Used for animation.
    
    **Type**: `int`.
    """
    LANDMARKS_3D_MEDIAPIPE_FACE = auto()
    """
    MediaPipe dense face landmarks in 3D. Each landmark is given by three coordinates (x,y,z) in camera space.

    **Type**: `Dict[InstanceId, ndarray[float32]]`. **Shape**: `(468, 3)`.
    """
    LANDMARKS_3D_SAI = auto()
    """
    SAI dense face landmarks in 3D. Each landmark is given by three coordinates (x,y,z) in camera space.

    **Type**: `Dict[InstanceId, ndarray[float32]]`. **Shape**: `(4840, 3)`.
    """
    LANDMARKS_MEDIAPIPE_FACE = auto()
    """
    MediaPipe dense face landmarks. Each landmark is given by two coordinates (x,y) in pixels.
    
    **Type**: `Dict[InstanceId, ndarray[float32]]`. **Shape**: `(468, 2)`.
    """
    LANDMARKS_SAI = auto()
    """
    SAI dense face landmarks. Each landmark is given by two coordinates (x,y) in pixels.
    
    **Type**: `Dict[InstanceId, ndarray[float32]]`. **Shape**: `(4840, 2)`.
    """



"""Top-level package for synthesisai."""

__author__ = """Synthesis AI"""
__email__ = 'anish@synthesis.ai'
__version__ = '0.1.0'

from .dataset import Grouping, SaiDataset
from .modality import Modality



import copy
import json
import os
from enum import Enum
from importlib.util import find_spec
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

import numpy as np
import pandas as pd
from pkg_resources import parse_version

from synthesisai.data_types import Item, OutOfFrameLandmarkStrategy
from synthesisai.item_loader_factory import _ItemLoaderFactory
from synthesisai.modality import Modality

CAM_NAME_PREFIX = "cam_"
FRAME_NUM_PREFIX = "f_"


if TYPE_CHECKING:
    Base = Sequence[dict]
else:
    Base = Sequence


class Grouping(Enum):
    """
    Different modalities of grouping Synthesis AI dataset.
    """

    NONE = "NONE"
    """
    Each image is treated independently.

    The size of the dataset is #scenes * #cameras * #frames  
    (assuming the same number of the cameras/frames per scene).
    """
    SCENE = "SCENE"
    """
    Items with the same scene are grouped into the list.

    The size of the dataset is #scenes. Each element is a `List[Item] with the same `SCENE_ID`.
    """
    CAMERA = "CAMERA"
    """
    Items with the same camera are grouped into the list.

    The size of the dataset is #cameras. Each element is a `List[Item] with the same `CAMERA_NAME`.
    """
    SCENE_CAMERA = "SCENE_CAMERA"
    """
    Items are grouped first by camera and then by scene.
    List of frames for a particular scene is indexed by `scene_id`.
    
    The size of the dataset is **#scenes** * **#cameras** , 
    each element is a `List[Item]` is a list of consecutive frames for given scene and camera.
    """


class SaiDataset(Base):
    """Synthesis AI dataset.

    This class provides access to all the modalities available in Synthesis AI generated datasets.
    """

    BODY_SEGMENTATION_MAPPING = {
        "default": 0,
        "background": 0,
        "beard": 1,
        "body": 2,
        "brow": 3,
        "cheek_left": 4,
        "cheek_right": 5,
        "chin": 6,
        "clothing": 7,
        "ear_left": 8,
        "ear_right": 9,
        "eye_left": 10,
        "eye_right": 11,
        "eyelashes": 12,
        "eyelid": 13,
        "eyes": 14,
        "forehead": 15,
        "glasses": 16,
        "hair": 17,
        "head": 18,
        "headphones": 19,
        "headwear": 20,
        "jaw": 21,
        "jowl": 22,
        "lip_lower": 23,
        "lip_upper": 24,
        "mask": 25,
        "mouth": 26,
        "mouthbag": 27,
        "mustache": 28,
        "neck": 29,
        "nose": 30,
        "nose_outer": 31,
        "nostrils": 32,
        "shoulders": 33,
        "smile_line": 34,
        "teeth": 35,
        "temples": 36,
        "tongue": 37,
        "undereye": 38,
        "eyebrows": 89,
        "torso_lower_left": 40,
        "torso_lower_right": 41,
        "torso_mid_left": 42,
        "torso_mid_right": 43,
        "torso_upper_left": 44,
        "torso_upper_right": 45,
        "arm_lower_left": 46,
        "arm_lower_right": 47,
        "arm_upper_left": 48,
        "arm_upper_right": 49,
        "hand_left": 50,
        "hand_right": 51,
        "finger1_mid_bottom_left": 52,
        "finger1_mid_bottom_right": 53,
        "finger1_mid_left": 54,
        "finger1_mid_right": 55,
        "finger1_mid_top_left": 56,
        "finger1_mid_top_right": 57,
        "finger2_mid_bottom_left": 58,
        "finger2_mid_bottom_right": 59,
        "finger2_mid_left": 60,
        "finger2_mid_right": 61,
        "finger2_mid_top_left": 62,
        "finger2_mid_top_right": 63,
        "finger3_mid_bottom_left": 64,
        "finger3_mid_bottom_right": 65,
        "finger3_mid_left": 66,
        "finger3_mid_right": 67,
        "finger3_mid_top_left": 68,
        "finger3_mid_top_right": 69,
        "finger4_mid_bottom_left": 70,
        "finger4_mid_bottom_right": 71,
        "finger4_mid_left": 72,
        "finger4_mid_right": 73,
        "finger4_mid_top_left": 74,
        "finger4_mid_top_right": 75,
        "finger5_mid_bottom_left": 76,
        "finger5_mid_bottom_right": 77,
        "finger5_mid_left": 78,
        "finger5_mid_right": 79,
        "finger5_mid_top_left": 80,
        "finger5_mid_top_right": 81,
        "nails_left": 82,
        "nails_right": 83,
        "leg_lower_left": 84,
        "leg_lower_right": 85,
        "leg_upper_left": 86,
        "leg_upper_right": 87,
        "foot_left": 88,
        "foot_right": 89,
        "pupil_left": 90,
        "pupil_right": 91,
        "eyelashes_left": 92,
        "eyelashes_right": 93,
        "eyelid_left": 92,
        "eyelid_right": 93,
        "eyebrow_left": 94,
        "eyebrow_right": 95,
        "glasses_frame": 96,
        "glasses_lens_left": 97,
        "glasses_lens_right": 98,
        "undereye_left": 99,
        "undereye_right": 100,
        "sclera_left": 101,
        "sclera_right": 102,
        "cornea_left": 103,
        "cornea_right": 104,
    }
    """
    Default body segmentation mapping.
    """

    CLOTHING_SEGMENTATION_MAPPING = {
        "default": 0,
        "background": 0,
        "sling dress": 1,
        "long sleeve dress": 2,
        "long sleeve outerwear": 3,
        "trousers": 4,
        "short sleeve shirt": 5,
        "short sleeve dress": 6,
        "vest dress": 7,
        "long sleeve shirt": 8,
        "shorts": 9,
        "short sleeve outerwear": 10,
        "skirt": 11,
        "scarf": 12,
        "vest": 13,
        "hat": 14,
        "shoe": 15,
    }
    """
    Default clothing segmentation mapping
    """

    FACE_SEGMENTATION_CLASSES = [
        "brow",
        "cheek_left",
        "cheek_right",
        "chin",
        "eye_left",
        "eye_right",
        "eyelid_left",
        "eyelid_right",
        "eyes",
        "jaw",
        "jowl",
        "lip_lower",
        "lip_upper",
        "mouth",
        "mouthbag",
        "nose",
        "nose_outer",
        "nostrils",
        "smile_line",
        "teeth",
        "undereye",
        "eyelashes_left",
        "eyelashes_right",
        "eyebrow_left",
        "eyebrow_right",
        "undereye_left",
        "undereye_right",
    ]
    "Segmentation classes included in the face bounding box."

    def __init__(
        self,
        root: Union[str, os.PathLike],
        modalities: Optional[List[Modality]] = None,
        body_segmentation_mapping: Optional[Dict[str, int]] = None,
        clothing_segmentation_mapping: Optional[Dict[str, int]] = None,
        face_segmentation_classes: Optional[List[str]] = None,
        face_bbox_pad: int = 0,
        grouping: Grouping = Grouping.NONE,
        out_of_frame_landmark_strategy: OutOfFrameLandmarkStrategy = OutOfFrameLandmarkStrategy.IGNORE,
        transform: Optional[
            Callable[[Dict[Modality, Any]], Dict[Modality, Any]]
        ] = None,
    ) -> None:
        """
        Initializes FaceApiDataset from the data in :attr:`root` directory, loading listed :attr:`modalities`.
        All dataset files should be located in the root directory, which should either look like this::

            root
            ├── metadata.jsonl
                0.cam_default.f_1.exr
                0.cam_default.f_1.rgb.png
                0.cam_default.f_1.info.json
                0.cam_default.f_1.segments.png
                0.cam_default.f_1.alpha.tif
                0.cam_default.f_1.depth.tif
                0.cam_default.f_1.normals.tif
                1.cam_default.f_1.exr
                1.cam_default.f_1.rgb.png
                1.cam_default.f_1.info.json
                1.cam_default.f_1.segments.png
                1.cam_default.f_1.alpha.tif
                1.cam_default.f_1.depth.tif
                1.cam_default.f_1.normals.tif

        Or it should look like this::

            root
            ├── metadata.jsonl
                0.cam_default.f_1.exr
                0.cam_default.f_1.rgba.npz
                0.cam_default.f_1.image_metadata.json
                0.cam_default.f_1.body_segmentation.npz
                0.cam_default.f_1.depth.npz
                0.cam_default.f_1.surface_normals.npz
                1.cam_default.f_1.exr
                1.cam_default.f_1.rgba.npz
                1.cam_default.f_1.image_metadata.json
                1.cam_default.f_1.body_segmentation.npz
                1.cam_default.f_1.depth.npz
                1.cam_default.f_1.surface_normals.npz

        No extra files are allowed, but all files which are not needed to load modalities listed may be absent.

        For instance, if only `RGB` and `BODY_SEGMENTATION` modalities are needed, then only one of the following sets of files are enough::

            root
            ├── 0.cam_default.f_1.rgb.png
                0.cam_default.f_1.info.json
                0.cam_default.f_1.segments.png
                1.cam_default.f_1.rgb.png
                1.cam_default.f_1.info.json
                1.cam_default.f_1.segments.png

            OR

            root
            ├── 0.cam_default.f_1.rgba.npz
                0.cam_default.f_1.image_metadata.json
                0.cam_default.f_1.body_segmentation.npz
                1.cam_default.f_1.rgba.npz
                1.cam_default.f_1.image_metadata.json
                1.cam_default.f_1.body_segmentation.npz

        If any of the required files are absent for at least one image, or any redundant files are located in the directory, :class:`ValueError` is raised.

        To work with segment modalities, :attr:`body_segmentation_mapping` parameter is used.
        This shows how to map a segmentation class name to integer representation.
        For example, to work with `background` (0)/`face` (1)/`hair` (2)/`body` (3) segmentation it may look like this::

            body_segmentation_mapping = {
                'default': 0,
                'background': 0,
                'beard': 1,
                'body': 3,
                'brow': 1,
                'cheek_left': 1,
                'cheek_right': 1,
                'chin': 1,
                'clothing': 3,
                'ear_left': 1,
                'ear_right': 1,
                'eye_left': 1,
                'eye_right': 1,
                'eyelashes': 1,
                'eyelid': 1,
                'eyes': 1,
                'forehead': 1,
                'glasses': 0,
                'hair': 2,
                'head': 1,
                'headphones': 0,
                'headwear': 0,
                'jaw': 1,
                'jowl': 1,
                'lip_lower': 1,
                'lip_upper': 1,
                'mask': 0,
                'mouth': 1,
                'mouthbag': 1,
                'mustache': 1,
                'neck': 3,
                'nose': 1,
                'nose_outer': 1,
                'nostrils': 1,
                'shoulders': 3,
                'smile_line': 1,
                'teeth': 1,
                'temples': 1,
                'tongue': 1,
                'undereye': 1
            }

        In addition :attr:`transform` function may be provided. If it is not `None` it will be applied to modality dict after each :meth:`__get__` call.

        For example, to flip rgb image and its segmentation::

            def transform(item: Dict[Modality, Any]) -> Dict[Modality, Any]:
                ret = item.copy()
                ret[Modality.RGB] = flip(modalities[Modality.RGB])
                ret[Modality.BODY_SEGMENTATION] = flip(modalities[Modality.BODY_SEGMENTATION])
                return ret

        :param Union[str,bytes,os.PathLike] root: Dataset root. All image files (ex. `0.cam_default.f_1.rgb.png`) should be located directly in this directory.
        :param Optional[List[Modality]] modalities: List of modalities to load. If None all the modalities are loaded.
        :param Optional[Dict[str,int]] body_segmentation_mapping: Mapping from object names to segmentation id. If `None` :attr:`BODY_SEGMENTATION_MAPPING` mapping is used.
        :param Optional[List[str]] face_segmentation_classes: List of object names considered to incorporate a face. If `None` :attr:`FACE_SEGMENTATION_CLASSES` is used.
        :param int face_bbox_pad: Extra area in pixels to pad around height and width of face bounding box.
        :param Optional[Callable[[Dict[Modality,Any]],Dict[Modality,Any]]] transform: Additional transforms to apply to modalities.
        """
        if body_segmentation_mapping is None:
            body_segmentation_mapping = self.BODY_SEGMENTATION_MAPPING
        if clothing_segmentation_mapping is None:
            clothing_segmentation_mapping = self.CLOTHING_SEGMENTATION_MAPPING
        if modalities is None:
            modalities = list(Modality)
        self._body_segmentation_mapping = body_segmentation_mapping
        self._clothing_segmentation_mapping = clothing_segmentation_mapping
        if face_segmentation_classes is None:
            face_segmentation_classes = self.FACE_SEGMENTATION_CLASSES
        self._face_segmentation_classes = face_segmentation_classes
        self._face_bbox_pad = face_bbox_pad
        self._modalities = sorted(modalities, key=lambda x: x.value)

        self._root = Path(root)
        if not (
            self._root.exists() and self._root.is_dir() and len(os.listdir(root)) > 0
        ):
            raise ValueError(
                f"{root} directory either doesn't exist or is not a directory or doesn't have files"
            )
        self._grouping = grouping

        metadata_records = []
        version_checked = False
        for file_path in self._root.glob("*"):
            if (
                file_path.name == "metadata.jsonl"
                or file_path.suffix == ".exr"
                or file_path.name.endswith("DS_Store")
            ):
                continue
            [
                scene_id,
                camera,
                frame,
                modality_name,
                modality_extension,
            ] = file_path.name.split(".")
            if not version_checked and modality_name == "info":
                with open(file_path, "r") as f:
                    info = json.load(f)
                dataset_version = info["version"]
                if parse_version(dataset_version) != parse_version("1.5"):
                    raise ValueError(
                        f"The version of this dataset is {dataset_version} which is not compatible with current face_api_dataset version. You could use an earlier version of face_api_dataset(<=1.0.4)."
                    )
                version_checked = True
            frame_num = int(frame.split(FRAME_NUM_PREFIX)[-1])
            cam_name = camera.split(CAM_NAME_PREFIX)[-1]
            extension = f"{modality_name}.{modality_extension}"
            record = {
                "SCENE_ID": int(scene_id),
                "CAMERA": camera,
                "CAMERA_NAME": cam_name,
                "FRAME": frame,
                "FRAME_NUM": int(frame_num),
                "EXTENSION": extension,
                "file_path": str(file_path),
            }

            if not scene_id.isdigit():
                raise ValueError(f"Unexpected file {file_path} in the dataset")

            metadata_records.append(record)

        self._item_loader = _ItemLoaderFactory.get_item_loader(
            self._root,
            self._modalities,
            metadata_records,
            out_of_frame_landmark_strategy,
            self._body_segmentation_mapping,
            self._clothing_segmentation_mapping,
            self._face_segmentation_classes,
            self._face_bbox_pad,
        )
        self._metadata = pd.DataFrame.from_records(metadata_records)

        if grouping is Grouping.NONE:
            group_columns = ["SCENE_ID", "CAMERA_NAME", "FRAME_NUM"]
        elif grouping is Grouping.SCENE:
            group_columns = ["SCENE_ID"]
        elif grouping is Grouping.CAMERA:
            group_columns = ["CAMERA_NAME"]
        elif grouping is Grouping.SCENE_CAMERA:
            group_columns = ["SCENE_ID", "CAMERA_NAME"]
        else:
            raise ValueError(f"Invalid grouping parameter {grouping}")

        self._metadata.set_index(
            ["SCENE_ID", "CAMERA_NAME", "FRAME_NUM"], inplace=True, drop=False
        )
        self._metadata.sort_index(inplace=True)
        self._group_columns = group_columns
        self._group_meta = self._metadata.groupby(level=group_columns)
        self._group_index: List[tuple] = list(self._group_meta.indices.keys())

        self._out_of_frame_landmark_strategy = out_of_frame_landmark_strategy
        self._transform = transform

    @property
    def body_segmentation_mapping(self) -> Dict[str, int]:
        """
        Segment mapping for the dataset.

        :type: Dict[str, int]
        """
        return self._body_segmentation_mapping

    @property
    def modalities(self) -> List[Modality]:
        """
        List of the loaded modalities.

        :type: List[Modality]
        """
        return self._modalities

    def __len__(self) -> int:
        return len(self._group_meta)

    @overload
    def __getitem__(self, index: int) -> dict:
        pass

    @overload
    def __getitem__(self, index: slice) -> "SaiDataset":
        pass

    def __getitem__(self, i: Union[int, slice]) -> Union[dict, "SaiDataset"]:
        if isinstance(i, slice):
            ret = copy.copy(self)
            ret._group_index = ret._group_index[i]
            indices = np.concatenate(
                [self._group_meta.indices[idx] for idx in ret._group_index]
            )
            ret._metadata = self._metadata.iloc[indices]
            ret._group_meta = ret._metadata.groupby(level=self._group_columns)
            return ret
        else:
            if self._transform is None:
                return self._get(i)
            else:
                return self._transform(self._get(i))

    def get_group_index(self) -> pd.DataFrame:
        if self._grouping is self._grouping.NONE:
            raise ValueError(
                "Group index is unavailable when groping is set to Grouping.NONE"
            )
        return pd.DataFrame(self._group_index, columns=self._group_columns)

    def _get(self, i: int) -> Union[Item, List[Item]]:
        if i > len(self):
            raise ValueError(f"Index {i} is out of bounds")
        group_idx = self._group_index[i]
        group_meta: pd.DataFrame = self._group_meta.get_group(group_idx)

        if self._grouping is Grouping.NONE:
            res = self._item_loader.get_item(group_idx, group_meta)
        elif self._grouping is Grouping.SCENE:
            scene_id = group_idx
            res = []
            for cam in group_meta.CAMERA_NAME.unique():
                cam_group_meta = group_meta[group_meta.CAMERA_NAME == cam]
                for frame in group_meta.FRAME_NUM.unique():
                    item_meta = cam_group_meta[cam_group_meta.FRAME_NUM == frame]
                    element_idx = (scene_id, cam, frame)
                    res.append(self._item_loader.get_item(element_idx, item_meta))
        elif self._grouping is Grouping.CAMERA:
            cam = group_idx
            res = []
            for scene_id in group_meta.SCENE_ID.unique():
                scene_group_meta = group_meta[group_meta.SCENE_ID == scene_id]
                for frame in scene_group_meta.FRAME_NUM.unique():
                    item_meta = scene_group_meta[scene_group_meta.FRAME_NUM == frame]
                    element_idx = (scene_id, cam, frame)
                    res.append(self._item_loader.get_item(element_idx, item_meta))
        elif self._grouping is Grouping.SCENE_CAMERA:
            scene_id, cam = group_idx
            res = []
            for frame in group_meta.FRAME_NUM.unique():
                item_meta = group_meta[group_meta.FRAME_NUM == frame]
                element_idx = (scene_id, cam, frame)
                res.append(self._item_loader.get_item(element_idx, item_meta))
        else:
            raise ValueError(f"Invalid grouping {self._grouping}.")
        return res



from enum import Enum
from typing import Dict, Tuple

import numpy as np

Item = dict
InstanceId = int
LandmarkId = int
Landmark_2D = Tuple[float, float]
Landmark_3D = Tuple[float, float, float]

class OutOfFrameLandmarkStrategy(Enum):
    IGNORE = "ignore"
    CLIP = "clip"

    @staticmethod
    def clip_landmarks_(
        landmarks: Dict[LandmarkId, Landmark_2D], height: int, width: int
    ) -> Dict[LandmarkId, Landmark_2D]:
        clipped_landmarks: Dict[LandmarkId, Landmark_2D] = {}
        for name, (x, y) in landmarks.items():
            xx = np.clip(x, 0, width - 1)
            yy = np.clip(y, 0, height - 1)
            clipped_landmarks[name] = (float(xx), float(yy))
        return clipped_landmarks



import json
import logging
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from synthesisai.data_types import (
    InstanceId,
    Item,
    Landmark_2D,
    Landmark_3D,
    LandmarkId,
    OutOfFrameLandmarkStrategy,
)
from synthesisai.item_loader import N_IBUG68_LANDMARKS, _ItemLoader
from synthesisai.modality import Modality

logger = logging.getLogger(__name__)


class _Extension(str, Enum):
    INFO = "image_metadata.json"
    RGB = "rgb.npz"
    NORMALS = "surface_normals.npz"
    DEPTH = "depth.npz"
    BODY_SEGMENTATION = "body_segmentation.npz"
    CLOTHING_SEGMENTATION = "clothing_segmentation.npz"
    INSTANCE_SEGMENTATION = "instance_segmentation.npz"


def _modality_files(modality: Modality) -> List[_Extension]:
    return {
        Modality.SCENE_ID: [_Extension.INFO],
        Modality.RGB: [_Extension.RGB],
        Modality.NORMALS: [_Extension.NORMALS],
        Modality.DEPTH: [_Extension.DEPTH],
        Modality.BODY_SEGMENTATION: [_Extension.INFO, _Extension.BODY_SEGMENTATION],
        Modality.CLOTHING_SEGMENTATION: [_Extension.INFO, _Extension.CLOTHING_SEGMENTATION],
        Modality.INSTANCE_SEGMENTATION: [_Extension.INSTANCE_SEGMENTATION],
        Modality.LANDMARKS_IBUG68: [_Extension.INFO],
        Modality.LANDMARKS_CONTOUR_IBUG68: None,
        Modality.LANDMARKS_KINECT_V2: [_Extension.INFO],
        Modality.LANDMARKS_COCO: [_Extension.INFO],
        Modality.LANDMARKS_MEDIAPIPE: [_Extension.INFO],
        Modality.LANDMARKS_MPEG4: [_Extension.INFO],
        Modality.LANDMARKS_3D_IBUG68: [_Extension.INFO],
        Modality.LANDMARKS_3D_KINECT_V2: [_Extension.INFO],
        Modality.LANDMARKS_3D_COCO: [_Extension.INFO],
        Modality.LANDMARKS_3D_MEDIAPIPE: [_Extension.INFO],
        Modality.LANDMARKS_3D_MPEG4: [_Extension.INFO],
        Modality.PUPILS: [_Extension.INFO],
        Modality.PUPILS_3D: [_Extension.INFO],
        Modality.IDENTITY: [_Extension.INFO],
        Modality.IDENTITY_METADATA: [_Extension.INFO],
        Modality.HAIR: [_Extension.INFO],
        Modality.FACIAL_HAIR: None,
        Modality.EXPRESSION: [_Extension.INFO],
        Modality.GAZE: [_Extension.INFO],
        Modality.FACE_BBOX: [
            _Extension.INFO,
            _Extension.BODY_SEGMENTATION,
            _Extension.INSTANCE_SEGMENTATION,
        ],
        Modality.CAM_TO_HEAD: [_Extension.INFO],
        Modality.HEAD_TO_CAM: [_Extension.INFO],
        Modality.WORLD_TO_HEAD: [_Extension.INFO],
        Modality.HEAD_TO_WORLD: [_Extension.INFO],
        Modality.CAM_TO_WORLD: [_Extension.INFO],
        Modality.WORLD_TO_CAM: [_Extension.INFO],
        Modality.CAM_INTRINSICS: [_Extension.INFO],
        Modality.CAMERA_NAME: [_Extension.INFO],
        Modality.FRAME_NUM: [_Extension.INFO],
        Modality.LANDMARKS_3D_MEDIAPIPE_FACE: None,
        Modality.LANDMARKS_3D_SAI: None,
        Modality.LANDMARKS_MEDIAPIPE_FACE: None,
        Modality.LANDMARKS_SAI: None,
    }[modality]


class _ItemLoaderV2(_ItemLoader):
    def __init__(
        self,
        root: Path,
        modalities: List[Modality],
        metadata_records: List[Dict[str, Union[int, str]]],
        out_of_frame_landmark_strategy: OutOfFrameLandmarkStrategy,
        body_segmentation_mapping: Dict[str, int],
        clothing_segmentation_mapping: Dict[str, int],
        face_segmentation_classes: List[str],
        face_bbox_pad: int,
    ):
        super().__init__(
            root,
            modalities,
            metadata_records,
            out_of_frame_landmark_strategy,
            body_segmentation_mapping,
            clothing_segmentation_mapping,
            face_segmentation_classes,
            face_bbox_pad,
        )
        filtered_modalities = []
        for m in self._modalities:
            if _modality_files(m) is None:
                logger.warn(f"Ignoring unsupported modality {m}")
            else:
                filtered_modalities.append(m)
        self._modalities = filtered_modalities
        self._needs_info = False
        for modality in self._modalities:
            if _Extension.INFO in _modality_files(modality):
                self._needs_info = True

        for record in metadata_records:
            scene_id = record["SCENE_ID"]
            camera = record["CAMERA"]
            frame = record["FRAME"]

            for modality in self._modalities:
                for extension in _modality_files(modality):
                    expected_file = (
                        root / f"{scene_id}.{camera}.{frame}.{extension.value}"
                    )
                    if not expected_file.exists():
                        raise ValueError(
                            f"Can't find file '{expected_file}' required for {modality.name} modality."
                        )

    def get_item(
        self, element_idx: Tuple[int, str, int], item_meta: pd.DataFrame
    ) -> Item:
        info = None
        if self._needs_info:
            info_file = Path(
                item_meta[item_meta.EXTENSION == _Extension.INFO].file_path.iloc[0]
            )
            with info_file.open("r") as f:
                info = json.load(f)

        ret: Item = {}
        for modality in self._modalities:
            ret[modality] = self._open_modality(modality, item_meta, element_idx, info)
        return ret

    def _open_modality(
        self,
        modality: Modality,
        item_meta: pd.DataFrame,
        element_idx: Tuple[int, str, int],  # (scene id, camera name, frame number)
        info: Optional[dict],
    ) -> Any:
        scene_id = item_meta.SCENE_ID.iloc[0]
        if modality == Modality.SCENE_ID:
            return scene_id

        if modality == Modality.RGB:
            file_path = item_meta[item_meta.EXTENSION == _Extension.RGB].file_path.iloc[
                0
            ]
            return self._read_rgb(file_path, element_idx)

        if modality == Modality.BODY_SEGMENTATION:
            file_path = item_meta[
                item_meta.EXTENSION == _Extension.BODY_SEGMENTATION
            ].file_path.iloc[0]
            segment_img, _ = self._read_body_segmentation(file_path, element_idx, info)
            return segment_img
        
        if modality == Modality.CLOTHING_SEGMENTATION:
            file_path = item_meta[item_meta.EXTENSION == _Extension.CLOTHING_SEGMENTATION].file_path.iloc[0]
            segment_img, _ = self._read_clothing_segmentation(file_path, element_idx, info)
            return segment_img

        if modality == Modality.INSTANCE_SEGMENTATION:
            file_path = item_meta[
                item_meta.EXTENSION == _Extension.INSTANCE_SEGMENTATION
            ].file_path.iloc[0]
            segment_img = self._read_instance_segmentation(file_path, element_idx)
            return segment_img

        if modality == Modality.NORMALS:
            normals_file = item_meta[
                item_meta.EXTENSION == _Extension.NORMALS
            ].file_path.iloc[0]
            img = np.load(normals_file)["image"]
            if element_idx in self._image_sizes:
                if self._image_sizes[element_idx] != img.shape[1::-1]:
                    raise ValueError(
                        f"Dimensions of different image modalities do not match for SCENE_ID={scene_id}"
                    )
            else:
                self._image_sizes[element_idx] = img.shape[1::-1]
            return img

        if modality == Modality.DEPTH:
            depth_file = item_meta[
                item_meta.EXTENSION == _Extension.DEPTH
            ].file_path.iloc[0]
            img = np.load(depth_file)["image"]
            if element_idx in self._image_sizes:
                if self._image_sizes[element_idx] != img.shape[1::-1]:
                    raise ValueError(
                        f"Dimensions of different image modalities do not match for SCENE_ID={scene_id}"
                    )
            else:
                self._image_sizes[element_idx] = img.shape[1::-1]

            return img

        if modality in (Modality.LANDMARKS_IBUG68, Modality.LANDMARKS_CONTOUR_IBUG68):
            return self._read_face_landmarks_2d(info, modality, element_idx)

        if modality in (
            Modality.LANDMARKS_KINECT_V2,
            Modality.LANDMARKS_COCO,
            Modality.LANDMARKS_MEDIAPIPE,
            Modality.LANDMARKS_MPEG4,
        ):
            landmark_meta = self._read_body_landmarks(
                info, modality, self._image_sizes[element_idx]
            )
            return landmark_meta

        if modality in (
            Modality.LANDMARKS_3D_IBUG68,
            Modality.LANDMARKS_3D_KINECT_V2,
            Modality.LANDMARKS_3D_COCO,
            Modality.LANDMARKS_3D_MEDIAPIPE,
            Modality.LANDMARKS_3D_MPEG4,
        ):
            landmark_meta = self._read_landmarks_3d(info, modality)
            return landmark_meta

        if modality == Modality.PUPILS:
            if element_idx not in self._image_sizes:
                raise ValueError(
                    "Pupils can only be loaded with at least one image modality"
                )
            scale = self._image_sizes[element_idx]

            pupils = {}
            for human in info["humans"]:
                human_pupils = [
                    human["pupil_coordinates"]["pupil_left"]["screen_space_pos"],
                    human["pupil_coordinates"]["pupil_right"]["screen_space_pos"],
                ]

                pupils_np = np.array(human_pupils, dtype=np.float64) * scale
                human_pupils = {
                    "pupil_left": tuple(pupils_np[0]),
                    "pupil_right": tuple(pupils_np[1]),
                }

                pupils[human["instance_id"]] = human_pupils

            return pupils

        if modality == Modality.PUPILS_3D:
            pupils = {}
            for human in info["humans"]:
                human_pupils = {
                    "pupil_left": tuple(
                        human["pupil_coordinates"]["pupil_left"]["camera_space_pos"]
                    ),
                    "pupil_right": tuple(
                        human["pupil_coordinates"]["pupil_right"]["camera_space_pos"]
                    ),
                }
                pupils[human["instance_id"]] = human_pupils

            return pupils

        if modality == Modality.IDENTITY:
            identities = {
                human["instance_id"]: human["preset_id"] for human in info["humans"]
            }
            return identities

        if modality == Modality.IDENTITY_METADATA:
            metadata = {
                human["instance_id"]: human["identity_metadata"]
                for human in info["humans"]
            }
            return metadata

        if modality == Modality.HAIR:
            hair = {}

            for human in info["humans"]:
                hair[human["instance_id"]] = {
                    "style": human["identity_metadata"]["hair_style"]
                }

            return hair

        if modality == Modality.EXPRESSION:
            expression = {}

            for human in info["humans"]:
                expression[human["instance_id"]] = human["pose"]["expression"]

            return expression

        if modality == Modality.GAZE:
            gaze = {}

            for human in info["humans"]:
                gaze[human["instance_id"]] = {
                    "right": human["gaze_values"]["eye_right"]["gaze_vector"],
                    "left": human["gaze_values"]["eye_left"]["gaze_vector"],
                }

            return gaze

        if modality == Modality.FACE_BBOX:
            body_segmentation_file_path = item_meta[
                item_meta.EXTENSION == _Extension.BODY_SEGMENTATION
            ].file_path.iloc[0]
            (
                body_segmentation,
                body_segmentation_mapping_int,
            ) = self._read_body_segmentation(
                body_segmentation_file_path, element_idx, info
            )

            body_segmentation_mapping = info["body_segmentation_mapping"]
            face_segmentation_classes = set(
                self._face_segmentation_classes
            ).intersection(set(body_segmentation_mapping.keys()))
            face_seg_idxs = [
                body_segmentation_mapping_int[body_segmentation_mapping[s]]
                for s in face_segmentation_classes
            ]
            face_mask = np.isin(body_segmentation, face_seg_idxs).astype(np.uint16)

            def get_bbox(img: np.ndarray):
                yxs = np.where(img != 0)
                if len(yxs[0]) == 0:
                    return (-1, -1, -1, -1)
                bbox = np.min(yxs[1]), np.min(yxs[0]), np.max(yxs[1]), np.max(yxs[0])
                return bbox

            height, width = self._image_sizes[element_idx]

            def expand_bbox(bbox: Tuple[int, int, int, int], padding: int):
                x0, y0, x1, y1 = bbox
                x0 = max(0, x0 - padding)
                y0 = max(0, y0 - padding)
                x1 = min(x1 + padding, width)
                y1 = min(y1 + padding, height)

                return (x0, y0, x1, y1)

            instance_segmentation_file_path = item_meta[
                item_meta.EXTENSION == _Extension.INSTANCE_SEGMENTATION
            ].file_path.iloc[0]
            instance_segmentation = self._read_instance_segmentation(
                instance_segmentation_file_path, element_idx
            )

            face_bboxes = {}

            for human in info["humans"]:
                instance_id = human["instance_id"]
                human_face_mask = np.copy(face_mask)
                # we only care about the face mask for this particular instance
                human_face_mask[instance_segmentation != instance_id] = 0
                face_bbox = get_bbox(human_face_mask)
                expanded_face_bbox = expand_bbox(face_bbox, self._face_bbox_pad)

                face_bboxes[instance_id] = expanded_face_bbox

            return face_bboxes

        if modality == Modality.CAM_TO_WORLD:
            return np.array(
                info["camera"]["transform_cam2world"]["mat_4x4"], dtype=np.float64
            )

        if modality == Modality.WORLD_TO_CAM:
            return np.array(
                info["camera"]["transform_world2cam"]["mat_4x4"], dtype=np.float64
            )

        if modality == Modality.HEAD_TO_CAM:
            head2cam = {}

            for human in info["humans"]:
                head2cam[human["instance_id"]] = np.array(
                    human["head_transform"]["transform_head2cam"]["mat_4x4"],
                    dtype=np.float64,
                )

            return head2cam

        if modality == Modality.CAM_TO_HEAD:
            head2cam = self._open_modality(
                Modality.HEAD_TO_CAM, item_meta, element_idx, info
            )
            cam2head = {}
            for instance_id, h2c in head2cam.items():
                cam2head[instance_id] = np.linalg.inv(h2c)

            return cam2head

        if modality == Modality.HEAD_TO_WORLD:
            head2world = {}

            for human in info["humans"]:
                head2world[human["instance_id"]] = np.array(
                    human["head_transform"]["transform_head2world"]["mat_4x4"],
                    dtype=np.float64,
                )

            return head2world

        if modality == Modality.WORLD_TO_HEAD:
            head2world = self._open_modality(
                Modality.HEAD_TO_WORLD, item_meta, element_idx, info
            )
            world2head = {}
            for instance_id, h2w in head2world.items():
                world2head[instance_id] = np.linalg.inv(h2w)

            return world2head

        if modality == Modality.CAM_INTRINSICS:
            return np.array(info["camera"]["intrinsics"]["mat_3x3"], dtype=np.float64)

        if modality == Modality.CAMERA_NAME:
            return item_meta.CAMERA_NAME.iloc[0]

        if modality == Modality.FRAME_NUM:
            return item_meta.FRAME_NUM.iloc[0]

        raise ValueError("Unknown modality")

    def _read_body_segmentation(
        self, body_segmentation_file: str, element_idx: tuple, info: Dict[str, Any]
    ) -> Tuple[np.ndarray, np.ndarray]:
        img = np.load(body_segmentation_file)["segmentation"]
        if element_idx in self._image_sizes:
            if self._image_sizes[element_idx] != img.shape[1::-1]:
                raise ValueError(
                    f"Dimensions of different image modalities do not match."
                )
        else:
            self._image_sizes[element_idx] = img.shape[1::-1]

        body_segmentation_mapping = info["body_segmentation_mapping"]
        body_segmentation_mapping_int = np.full(
            np.max(list(body_segmentation_mapping.values())) + 1,
            self._body_segmentation_mapping["default"],
            dtype=np.uint8,
        )
        for key, value in body_segmentation_mapping.items():
            if key in self._body_segmentation_mapping:
                body_segmentation_mapping_int[value] = self._body_segmentation_mapping[
                    key
                ]

        body_segmentation = body_segmentation_mapping_int[img]
        return body_segmentation, body_segmentation_mapping_int
    
    def _read_clothing_segmentation(
        self, clothing_segmentation_file: str, element_idx: tuple, info: Dict[str, Any]
    ) -> Tuple[np.ndarray, np.ndarray]:
        img = np.load(clothing_segmentation_file)["segmentation"]
        if element_idx in self._image_sizes:
            if self._image_sizes[element_idx] != img.shape[1::-1]:
                raise ValueError(
                    f"Dimensions of different image modalities do not match."
                )
        else:
            self._image_sizes[element_idx] = img.shape[1::-1]

        clothing_segmentation_mapping = info["clothing_segmentation_mapping"]
        clothing_segmentation_mapping_int = np.full(
            np.max(list(clothing_segmentation_mapping.values())) + 1,
            self._clothing_segmentation_mapping["default"],
            dtype=np.uint8,
        )
        for key, value in clothing_segmentation_mapping.items():
            if key in self._clothing_segmentation_mapping:
                clothing_segmentation_mapping_int[value] = self._clothing_segmentation_mapping[
                    key
                ]

        clothing_segmentation = clothing_segmentation_mapping_int[img]
        return clothing_segmentation, clothing_segmentation_mapping_int

    def _read_instance_segmentation(
        self, instance_segmentation_file: str, element_idx: tuple
    ) -> np.ndarray:
        img = np.load(instance_segmentation_file)["segmentation"]
        if element_idx in self._image_sizes:
            if self._image_sizes[element_idx] != img.shape[1::-1]:
                raise ValueError(
                    f"Dimensions of different image modalities do not match."
                )
        else:
            self._image_sizes[element_idx] = img.shape[1::-1]

        return img

    def _read_rgb(self, rgb_file: str, element_idx: tuple) -> np.ndarray:
        img = np.load(rgb_file)["image"]
        if element_idx in self._image_sizes:
            if self._image_sizes[element_idx] != img.shape[1::-1]:
                raise ValueError("Dimensions of different modalities do not match")
        else:
            self._image_sizes[element_idx] = img.shape[1::-1]
        if img.shape[2] != 3:
            raise ValueError(
                f"Expected RGBA image to have 3 channels, got {img.shape[2]}"
            )

        return img

    @staticmethod
    def _read_modality_meta(info: dict, mdt: Modality) -> dict:
        if mdt in (Modality.LANDMARKS_IBUG68, Modality.LANDMARKS_3D_IBUG68):
            meta_dict = {
                human["instance_id"]: human["landmarks"]["ibug68"]
                for human in info["humans"]
            }
        elif mdt is Modality.LANDMARKS_CONTOUR_IBUG68:
            meta_dict = {
                human["instance_id"]: human["landmarks"]["ibug68_contour"]
                for human in info["humans"]
            }
        elif mdt in (Modality.LANDMARKS_COCO, Modality.LANDMARKS_3D_COCO):
            meta_dict = {
                human["instance_id"]: human["landmarks"]["coco_whole_body"]
                for human in info["humans"]
            }
        elif mdt in (Modality.LANDMARKS_KINECT_V2, Modality.LANDMARKS_3D_KINECT_V2):
            meta_dict = {
                human["instance_id"]: human["landmarks"]["kinect_v2"]
                for human in info["humans"]
            }
        elif mdt in (Modality.LANDMARKS_MEDIAPIPE, Modality.LANDMARKS_3D_MEDIAPIPE):
            meta_dict = {
                human["instance_id"]: human["landmarks"]["mediapipe"]["body"]
                for human in info["humans"]
            }
        elif mdt in (Modality.LANDMARKS_MPEG4, Modality.LANDMARKS_3D_MPEG4):
            meta_dict = {
                human["instance_id"]: human["landmarks"]["mpeg4_fba"]
                for human in info["humans"]
            }
        else:
            raise ValueError(f"Unrecognized modality for 3D landmarks {mdt}.")
        return meta_dict

    @classmethod
    def _read_body_landmarks(
        cls, info: dict, mdt: Modality, image_size: Tuple[int, int]
    ) -> Dict[InstanceId, Dict[LandmarkId, Landmark_2D]]:
        meta_dict = cls._read_modality_meta(info, mdt)
        result = {}

        for instance_id, landmarks in meta_dict.items():
            lm = {}
            for item in landmarks:
                idd = item["id"]
                x, y = item["screen_space_pos"]
                w, h = image_size
                lm[int(idd)] = (x * w, y * h)
            result[instance_id] = lm

        return result

    def _read_face_landmarks_2d(
        self, info: dict, mdt: Modality, element_idx: tuple
    ) -> Dict[InstanceId, Dict[LandmarkId, Landmark_2D]]:
        if element_idx not in self._image_sizes:
            raise ValueError(
                "Landmarks can only be loaded with at least one image modality"
            )

        w, h = self._image_sizes[element_idx]
        meta_dict = self._read_modality_meta(info, mdt)
        landmarks_2d: Dict[InstanceId, Dict[LandmarkId, Landmark_2D]] = {}

        for instance_id, landmarks in meta_dict.items():
            if mdt is Modality.LANDMARKS_IBUG68:
                n_landmarks = len(landmarks)
                if n_landmarks != N_IBUG68_LANDMARKS:
                    msg = f"Error reading landmarks for item. Got {n_landmarks} landmarks instead of {N_IBUG68_LANDMARKS}"
                    raise ValueError(msg)

            landmarks_2d[instance_id] = {}

            for landmark in landmarks:
                x, y = landmark["screen_space_pos"]
                x, y = x * w, y * h
                landmarks_2d[instance_id][int(landmark["id"])] = (x, y)

            if self._out_of_frame_landmark_strategy is OutOfFrameLandmarkStrategy.CLIP:
                landmarks_2d[instance_id] = OutOfFrameLandmarkStrategy.clip_landmarks_(
                    landmarks_2d[instance_id], h, w
                )

        return landmarks_2d

    @classmethod
    def _read_landmarks_3d(
        cls, info: dict, mdt: Modality
    ) -> Dict[LandmarkId, Landmark_3D]:
        meta_dict = cls._read_modality_meta(info, mdt)
        landmarks_3d: Dict[InstanceId, Dict[LandmarkId, Landmark_3D]] = {}

        for instance_id, landmarks in meta_dict.items():
            landmarks_3d[instance_id] = {}
            for landmark in landmarks:
                lmk_name = landmark["id"]
                landmarks_3d[instance_id][int(lmk_name)] = tuple(
                    landmark["camera_space_pos"]
                )

        return landmarks_3d
